package parser;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.TreeMap;

import utils.Fraction;

public final class LPParser {

	private final InputStream is;

	private final Map<String, Variable> vars;

	private boolean eof = false;

	private char buffer;

	private String token;

	public LPParser(InputStream is) throws IOException {
		this.is = is;
		vars = new TreeMap<String, Variable>();

		nextChar();
		nextToken();
	}

	public LPProgram readProgram() throws IOException {
		Command cmd = readCommand();
		LinearFormula z = readFormular();
		Collection<Constraint> constraints = new ArrayList<Constraint>();
		while (readNewLine() && !eof) {
			nextToken();
			Constraint c = readContraint();
			constraints.add(c);
		}

		return new LPProgram(cmd, z, constraints);
	}

	private Command readCommand() throws IOException {
		if (token.equals("min")) {
			nextToken();
			return Command.MIN;
		} else if (token.equals("max")) {
			nextToken();
			return Command.MAX;
		}

		return null;
	}

	private Constraint readContraint() throws IOException {
		LinearFormula a = readFormular();
		Comparator cmp = readComperator();
		nextToken();
		Constant b = readConstant();

		return new Constraint(a, cmp, b);
	}

	private Comparator readComperator() throws IOException {
		if (token.equals("<=")) {
			return Comparator.LESS_EQUAL;
		} else if (token.equals(">=")) {
			return Comparator.GREATER_EQUAL;
		} else if (token.equals("=")) {
			return Comparator.EQUAL;
		}

		return null;
	}

	private Constant readConstant() throws IOException {
		Fraction c = readFraction();
		return new Constant(c);
	}

	private Fraction readFraction() throws IOException {
		String t = token;
		nextToken();
		return new Fraction(Integer.valueOf(t), 1);
	}

	private boolean readNewLine() throws IOException {
		if (token.startsWith("\n")) {
			return true;
		}

		return false;
	}

	private LinearFormula readFormular() throws IOException {
		Collection<Term> terms = new ArrayList<Term>();
		while (!readNewLine() && readComperator() == null) {
			Operator op = readOperator();
			Term t = readTerm(op == Operator.MINUS);
			terms.add(t);
		}
		return new LinearFormula(terms);
	}

	private Operator readOperator() throws IOException {
		if (token.equals("+")) {
			nextToken();
			return Operator.PLUS;
		} else if (token.equals("-")) {
			nextToken();
			return Operator.MINUS;
		}

		return null;
	}

	private Term readTerm(boolean negate) throws IOException {
		Fraction coefficient;
		if (Character.isDigit(token.charAt(0))) {
			coefficient = readFraction();
		} else {
			coefficient = Fraction.ONE;
		}
		Variable var = readVariable();

		if (negate) {
			coefficient = coefficient.negative();
		}

		return new Term(coefficient, var);
	}

	private Variable readVariable() throws IOException {
		String name = token;
		nextToken();

		if (vars.containsKey(name)) {
			return vars.get(name);
		}

		Variable var = new Variable(name);
		vars.put(name, var);

		return var;
	}

	private boolean isOperator(char ch) {
		return ch == '+' || ch == '-' || ch == '<' || ch == '>' || ch == '=';
	}

	private void nextToken() throws IOException {
		while (Character.isWhitespace(buffer) && buffer != '\n') {
			nextChar();
		}

		StringBuilder sb = new StringBuilder();

		if (Character.isDigit(buffer)) {
			do {
				sb.append(buffer);
				nextChar();
			} while (Character.isDigit(buffer));
		} else if (Character.isLetter(buffer)) {
			do {
				sb.append(buffer);
				nextChar();
			} while (Character.isLetter(buffer) || Character.isDigit(buffer));
		} else if (buffer == '\n') {
			sb.append(buffer);
			nextChar();
		} else if (isOperator(buffer)) {
			do {
				sb.append(buffer);
				nextChar();
			} while (isOperator(buffer));
		}

		token = sb.toString();
		if (token.length() == 0) {
			eof = true;
		}
	}

	private void nextChar() throws IOException {
		int b = is.read();

		if (b < 0) {
			eof = true;
			buffer = 0;
		} else {
			buffer = (char) b;
		}
	}
}
