package solver;

import java.io.*;
import java.util.*;

import parser.*;
import parser.Comparator;
import utils.Fraction;

public final class LPSolver {

	public static boolean help_simplex(Tableau tab) {
		Fraction[] helpCol = new Fraction[tab.cntRows()];
		for (int i = 0; i < helpCol.length - 2; i++) {
			helpCol[i] = Fraction.ONE.negative();
		}
		helpCol[helpCol.length - 2] = Fraction.ONE;
		helpCol[helpCol.length - 1] = Fraction.ZERO;

		String helpBasis = "_x0";
		tab.insertCol(0, helpBasis, helpCol);

		System.out.println(tab);

		System.out.println("Phase 1:");

		int valueRow = tab.cntRows() - 2;
		int constCol = tab.cntCols() - 1;

		int pivotCol = 0;
		int pivotRow = -1;

		Fraction maxTheta = Fraction.ZERO;

		for (int i = 0; i < valueRow; i++) {
			Fraction s = tab.getCell(i, constCol);
			Fraction theta = s.negative();
			if (theta.compareTo(maxTheta) > 0) {
				pivotRow = i;
				maxTheta = theta;
			}
		}

		assert (pivotRow >= 0);

		tab.addToBasis(pivotRow, pivotCol);

		System.out.println(tab);

		simplex(tab, true);

		pivotRow = tab.searchBasis(helpBasis);

		if (pivotRow >= 0) {
			if (tab.getCell(pivotRow, constCol).compareTo(Fraction.ZERO) > 0) {
				System.out.println("Hilfsproblem hat Optimum größer als 0 -> Ursprungsproblem ist unzulässig!");
				return false;
			}
			
			for (int col = 1; col < constCol; col++) {
				if (tab.getCell(pivotRow, col).compareTo(Fraction.ZERO) != 0) {
					pivotCol = col;
					break;
				}
			}
			
			tab.addToBasis(pivotRow, pivotCol);
			System.out.println(tab);
		}
		
		tab.delRow(tab.cntRows() - 2);
		tab.delCol(0);

		System.out.println(tab);
		return true;
	}

	protected static void simplex(Tableau tab, boolean help) {
		int valueRow = tab.cntRows() - 1;
		int constCol = tab.cntCols() - 1;

		if (help) {
			valueRow--;
		}

		while (true) {
			int pivotCol = 0;
			int pivotRow = -1;

			Fraction maxDecrease = tab.getCell(valueRow, pivotCol);

			for (int i = 1; i < constCol; i++) {
				Fraction decrese = tab.getCell(valueRow, i);
				if (decrese.compareTo(maxDecrease) < 0) {
					pivotCol = i;
					maxDecrease = decrese;
				}
			}

			if (maxDecrease.numerator() >= 0) {
				break;
			}

			Fraction minTheta = Fraction.ZERO;

			for (int i = 0; i < valueRow; i++) {
				Fraction r = tab.getCell(i, constCol);
				Fraction s = tab.getCell(i, pivotCol);
				if (s.numerator() > 0) {
					Fraction theta = r.dividedBy(s);
					if (pivotRow < 0 || theta.compareTo(minTheta) < 0) {
						pivotRow = i;
						minTheta = theta;
					}
				}
			}

			assert (pivotRow >= 0);

			tab.addToBasis(pivotRow, pivotCol);

			System.out.println(tab);
		}
	}

	public static void main(String[] args) {
		try {
			InputStream is;
			if (args.length > 0) {
				is = new FileInputStream(args[0]);
			} else {
				is = System.in;
			}

			LPParser parser = new LPParser(is);

			LPProgram lp = parser.readProgram();

			System.out.println(lp);

			TreeSet<String> varset = new TreeSet<String>();
			lp.getVarSet(varset);

			System.out.println(varset);

			Comparator[] signs = new Comparator[varset.size()];

			Tableau tab = new Tableau();

			{
				Fraction[] dummy = new Fraction[0];
				for (String var : varset) {
					tab.addCol(var, dummy);
				}
				tab.addCol("_b", dummy);
			}

			for (Constraint c : lp.constraints) {
				Fraction[] row = new Fraction[varset.size() + 1];

				for (int i = 0; i < row.length; i++) {
					row[i] = Fraction.ZERO;
				}

				if (c.a.terms.size() == 1 && c.b.value.numerator() == 0) {
					String x = c.a.terms.iterator().next().var.name;
					int index = varset.headSet(x).size();

					signs[index] = c.cmp;
				} else {
					for (Term t : c.a.terms) {
						Fraction a = t.coefficient;
						String x = t.var.name;
						int index = varset.headSet(x).size();

						row[index] = a;
					}

					row[varset.size()] = c.b.value;

					if (c.cmp == Comparator.LESS_EQUAL
							|| c.cmp == Comparator.EQUAL) {
						tab.addRow(row);
					}

					if (c.cmp == Comparator.GREATER_EQUAL
							|| c.cmp == Comparator.EQUAL) {
						Fraction[] row2 = new Fraction[row.length];
						for (int i = 0; i < row2.length; i++) {
							row2[i] = row[i].negative();
						}
						tab.addRow(row2);
					}
				}
			}

			int maxCols = tab.cntCols() + tab.cntRows();
			Fraction[] valueRow = new Fraction[maxCols];
			Fraction[] helpRow = new Fraction[maxCols];

			for (int i = 0; i < valueRow.length; i++) {
				valueRow[i] = Fraction.ZERO;
				helpRow[i] = Fraction.ZERO;
			}

			for (Term t : lp.z.terms) {
				Fraction a = t.coefficient;
				String x = t.var.name;
				int index = varset.headSet(x).size();

				valueRow[index] = a;
			}

			if (lp.cmd.equals(Command.MAX)) {
				for (int i = 0; i < valueRow.length - 1; i++) {
					valueRow[i] = valueRow[i].negative();
				}
			}

			System.out.println(Arrays.toString(signs));
			System.out.println();

			// TODO x <= 0 -> -x >= 0
			// TODO x <=> 0 -> x = x+ - x-, x+ >= 0, x- >= 0
			// TODO a >= b -> -a <= -b

			boolean help = false;

			for (int i = 0; i < tab.cntRows(); i++) {
				tab.addBasis();

				if (tab.getCell(i, tab.cntCols() - 1).compareTo(Fraction.ZERO) < 0) {
					help = true;
				}
			}

			if (help) {
				tab.addRow(helpRow);
				tab.addRow(valueRow);

				if (!help_simplex(tab)) {
					return;
				}
			} else {
				tab.addRow(valueRow);
				System.out.println(tab);
			}

			System.out.println("Phase 2:");
			simplex(tab, false);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

}
