package solver;

import utils.Fraction;

public final class Tableau {

	private String[] names = new String[0];
	private String[] basis = new String[0];
	private Fraction[][] matrix = new Fraction[0][0];
	
	public void addCol(String name, Fraction[] col) {
		if (matrix.length == 0) {
			matrix = new Fraction[col.length][];
		}

		String[] newNames = new String[names.length + 1];
		System.arraycopy(names, 0, newNames, 0, names.length);
		newNames[names.length] = name;
		names = newNames;
		
		for (int i = 0; i < col.length; i++) {
			Fraction[] newRow = new Fraction[matrix[i].length + 1];
			System.arraycopy(matrix[i], 0, newRow, 0, matrix[i].length);
			newRow[matrix[i].length] = col[i];
			matrix[i] = newRow;
		}
	}

	public void insertCol(int idxCol, String name, Fraction[] col) {
		addCol(name, col);
		moveCol(matrix[0].length - 1, idxCol);
	}

	private void moveCol(int from, int to) {
		assert (from > to);
		
		String s = names[from];
		int move = names.length - to - 1;
		System.arraycopy(names, to, names, to + 1, move);
		names[to] = s;
		
		for (int row = 0; row < matrix.length; row++) {
			Fraction tmp = matrix[row][from];
			System.arraycopy(matrix[row], to, matrix[row], to + 1, move);
			matrix[row][to] = tmp;
		}
	}

	public void delCol(int idxCol) {
		String[] newNames = new String[names.length - 1];
		System.arraycopy(names, 0, newNames, 0, idxCol);
		System.arraycopy(names, idxCol + 1, newNames, idxCol, newNames.length - idxCol);
		names = newNames;
		
		for (int row = 0; row < matrix.length; row++) {
			Fraction[] newRow = new Fraction[matrix[row].length - 1];
			System.arraycopy(matrix[row], 0, newRow, 0, idxCol);
			System.arraycopy(matrix[row], idxCol + 1, newRow, idxCol, newRow.length - idxCol);
			matrix[row] = newRow;
		}
	}

	public void addRow(Fraction[] row) {
		if (matrix.length == 0) {
			matrix = new Fraction[][] { row };
		} else {
			assert (row.length == matrix[0].length);
			Fraction[][] newMatrix = new Fraction[matrix.length + 1][];
			System.arraycopy(matrix, 0, newMatrix, 0, matrix.length);
			newMatrix[matrix.length] = row;
			matrix = newMatrix;
		}
	}

	public void delRow(int idxRow) {
		Fraction[][] newMatrix = new Fraction[matrix.length - 1][];
		System.arraycopy(matrix, 0, newMatrix, 0, idxRow);
		System.arraycopy(matrix, idxRow + 1, newMatrix, idxRow, newMatrix.length - idxRow);
		matrix = newMatrix;
	}

	public String toString() {
		int cols = 1 + cntCols();
		Table t = new Table(cols);
		String[] strings = new String[cols];
		
		strings[0] = "";
		System.arraycopy(names, 0, strings, 1, names.length);
		t.addLine(strings);
		
		for (int row = 0; row < matrix.length; row++) {
			strings[0] = row < basis.length ? basis[row] : "";
			for (int col = 0; col < matrix[row].length; col++) {
				strings[1 + col] = matrix[row][col].toString();
			}
			t.addLine(strings);
		}

		return t.toString();
	}

	public int cntRows() {
		return matrix.length;
	}

	public int cntCols() {
		return (matrix.length > 0) ? matrix[0].length : 0;
	}
	
	public Fraction getCell(int row, int col) {
		return matrix[row][col];
	}

	public void setCell(int row, int col, Fraction d) {
		matrix[row][col] = d;
	}

	public void addBasis() {
		int row = basis.length;
		
		Fraction[] slackCol = new Fraction[cntRows()];
		for (int j = 0; j < slackCol.length; j++) {
			slackCol[j] = Fraction.ZERO;
		}
		slackCol[row] = Fraction.ONE;
		
		String newBasisName = "_s" + row;
		
		String[] newBasis = new String[basis.length + 1];
		System.arraycopy(basis, 0, newBasis, 0, basis.length);
		newBasis[basis.length] = newBasisName;
		basis = newBasis;

		insertCol(cntCols() - 1, newBasisName, slackCol);
	}

	public Fraction addToBasis(int pivotRow, int pivotCol) {
		Fraction pivotValue = getCell(pivotRow, pivotCol);

		for (int i = 0; i < cntCols(); i++) {
			setCell(pivotRow, i, getCell(pivotRow, i).dividedBy(pivotValue));
		}

		for (int i = 0; i < cntRows(); i++) {
			if (i != pivotRow) {
				Fraction factor = getCell(i, pivotCol);
				for (int j = 0; j < cntCols(); j++) {
					Fraction sub = factor.times(getCell(pivotRow, j));
					setCell(i, j, getCell(i, j).minus(sub));
				}
			}
		}
		
		basis[pivotRow] = names[pivotCol];
		
		return pivotValue;
	}

	public int searchBasis(String b) {
		for (int i = 0; i < basis.length; i++) {
			if (basis[i].equals(b)) {
				return i;
			}
		}
		
		return -1;
	}

}
